@@ -1,16 +1,23 @@
 //! Main input logic.
 
-use zellij_utils::{termion, zellij_tile};
-
-use crate::{os_input_output::ClientOsApi, ClientInstruction, CommandIsExecuting};
 use zellij_utils::{
-    channels::{SenderWithContext, OPENCALLS},
-    errors::ContextType,
+    input::{
+        mouse::{MouseButton, MouseEvent},
+        options::Options,
+    },
+    termion, zellij_tile,
+};
+
+use crate::{
+    os_input_output::ClientOsApi, ClientInstruction, CommandIsExecuting, InputInstruction,
+};
+use zellij_utils::{
+    channels::{Receiver, SenderWithContext, OPENCALLS},
+    errors::{ContextType, ErrorContext},
     input::{actions::Action, cast_termion_key, config::Config, keybinds::Keybinds},
     ipc::{ClientToServerMsg, ExitReason},
 };
 
-use termion::input::TermReadEventsAndRaw;
 use zellij_tile::data::{InputMode, Key};
 
 /// Handles the dispatching of [`Action`]s according to the current
@@ -20,10 +27,12 @@ struct InputHandler {
     mode: InputMode,
     os_input: Box<dyn ClientOsApi>,
     config: Config,
+    options: Options,
     command_is_executing: CommandIsExecuting,
     send_client_instructions: SenderWithContext<ClientInstruction>,
     should_exit: bool,
     pasting: bool,
+    receive_input_instructions: Receiver<(InputInstruction, ErrorContext)>,
 }
 
 impl InputHandler {
@@ -32,17 +41,21 @@ impl InputHandler {
         os_input: Box<dyn ClientOsApi>,
         command_is_executing: CommandIsExecuting,
         config: Config,
+        options: Options,
         send_client_instructions: SenderWithContext<ClientInstruction>,
         mode: InputMode,
+        receive_input_instructions: Receiver<(InputInstruction, ErrorContext)>,
     ) -> Self {
         InputHandler {
             mode,
             os_input,
             config,
+            options,
             command_is_executing,
             send_client_instructions,
             should_exit: false,
             pasting: false,
+            receive_input_instructions,
         }
     }
 
@@ -54,18 +67,25 @@ impl InputHandler {
         let alt_left_bracket = vec![27, 91];
         let bracketed_paste_start = vec![27, 91, 50, 48, 48, 126]; // \u{1b}[200~
         let bracketed_paste_end = vec![27, 91, 50, 48, 49, 126]; // \u{1b}[201
+
+        if !self.options.disable_mouse_mode {
+            self.os_input.enable_mouse();
+        }
         loop {
             if self.should_exit {
                 break;
             }
-            let stdin_buffer = self.os_input.read_from_stdin();
-            for key_result in stdin_buffer.events_and_raw() {
-                match key_result {
-                    Ok((event, raw_bytes)) => match event {
+            match self.receive_input_instructions.recv() {
+                Ok((InputInstruction::KeyEvent(event, raw_bytes), _error_context)) => {
+                    match event {
                         termion::event::Event::Key(key) => {
                             let key = cast_termion_key(key);
                             self.handle_key(&key, raw_bytes);
                         }
+                        termion::event::Event::Mouse(me) => {
+                            let mouse_event = zellij_utils::input::mouse::MouseEvent::from(me);
+                            self.handle_mouse_event(&mouse_event);
+                        }
                         termion::event::Event::Unsupported(unsupported_key) => {
                             // we have to do this because of a bug in termion
                             // this should be a key event and not an unsupported event
@@ -74,21 +94,22 @@ impl InputHandler {
                                 self.handle_key(&key, raw_bytes);
                             } else if unsupported_key == bracketed_paste_start {
                                 self.pasting = true;
+                                self.handle_unknown_key(raw_bytes);
                             } else if unsupported_key == bracketed_paste_end {
                                 self.pasting = false;
+                                self.handle_unknown_key(raw_bytes);
                             } else {
                                 // this is a hack because termion doesn't recognize certain keys
                                 // in this case we just forward it to the terminal
                                 self.handle_unknown_key(raw_bytes);
                             }
                         }
-                        termion::event::Event::Mouse(_) => {
-                            // Mouse events aren't implemented yet,
-                            // use a NoOp untill then.
-                        }
-                    },
-                    Err(err) => panic!("Encountered read error: {:?}", err),
+                    }
                 }
+                Ok((InputInstruction::SwitchToMode(input_mode), _error_context)) => {
+                    self.mode = input_mode;
+                }
+                Err(err) => panic!("Encountered read error: {:?}", err),
             }
         }
     }
@@ -109,7 +130,7 @@ impl InputHandler {
                 self.dispatch_action(action);
             }
         } else {
-            for action in Keybinds::key_to_actions(&key, raw_bytes, &self.mode, keybinds) {
+            for action in Keybinds::key_to_actions(key, raw_bytes, &self.mode, keybinds) {
                 let should_exit = self.dispatch_action(action);
                 if should_exit {
                     self.should_exit = true;
@@ -117,6 +138,28 @@ impl InputHandler {
             }
         }
     }
+    fn handle_mouse_event(&mut self, mouse_event: &MouseEvent) {
+        match *mouse_event {
+            MouseEvent::Press(button, point) => match button {
+                MouseButton::WheelUp => {
+                    self.dispatch_action(Action::ScrollUpAt(point));
+                }
+                MouseButton::WheelDown => {
+                    self.dispatch_action(Action::ScrollDownAt(point));
+                }
+                MouseButton::Left => {
+                    self.dispatch_action(Action::LeftClick(point));
+                }
+                _ => {}
+            },
+            MouseEvent::Release(point) => {
+                self.dispatch_action(Action::MouseRelease(point));
+            }
+            MouseEvent::Hold(point) => {
+                self.dispatch_action(Action::MouseHold(point));
+            }
+        }
+    }
 
     /// Dispatches an [`Action`].
     ///
@@ -140,17 +183,20 @@ impl InputHandler {
                 should_break = true;
             }
             Action::SwitchToMode(mode) => {
+                // this is an optimistic update, we should get a SwitchMode instruction from the
+                // server later that atomically changes the mode as well
                 self.mode = mode;
                 self.os_input
                     .send_to_server(ClientToServerMsg::Action(action));
             }
             Action::CloseFocus
             | Action::NewPane(_)
-            | Action::NewTab
+            | Action::NewTab(_)
             | Action::GoToNextTab
             | Action::GoToPreviousTab
             | Action::CloseTab
             | Action::GoToTab(_)
+            | Action::ToggleTab
             | Action::MoveFocusOrTab(_) => {
                 self.command_is_executing.blocking_input_thread();
                 self.os_input
@@ -180,16 +226,24 @@ impl InputHandler {
 pub(crate) fn input_loop(
     os_input: Box<dyn ClientOsApi>,
     config: Config,
+    options: Options,
     command_is_executing: CommandIsExecuting,
     send_client_instructions: SenderWithContext<ClientInstruction>,
     default_mode: InputMode,
+    receive_input_instructions: Receiver<(InputInstruction, ErrorContext)>,
 ) {
     let _handler = InputHandler::new(
         os_input,
         command_is_executing,
         config,
+        options,
         send_client_instructions,
         default_mode,
+        receive_input_instructions,
     )
     .handle_input();
 }
+
+#[cfg(test)]
+#[path = "./unit/input_handler_tests.rs"]
+mod grid_tests;
