@@ -1,4 +1,4 @@
-use zellij_utils::input::actions::Action;
+use zellij_utils::pane_size::Size;
 use zellij_utils::{interprocess, libc, nix, signal_hook, termion, zellij_tile};
 
 use interprocess::local_socket::LocalSocketStream;
@@ -15,7 +15,6 @@ use zellij_tile::data::Palette;
 use zellij_utils::{
     errors::ErrorContext,
     ipc::{ClientToServerMsg, IpcReceiverWithContext, IpcSenderWithContext, ServerToClientMsg},
-    pane_size::PositionAndSize,
     shared::default_palette,
 };
 
@@ -35,7 +34,7 @@ fn unset_raw_mode(pid: RawFd, orig_termios: termios::Termios) {
     };
 }
 
-pub(crate) fn get_terminal_size_using_fd(fd: RawFd) -> PositionAndSize {
+pub(crate) fn get_terminal_size_using_fd(fd: RawFd) -> Size {
     // TODO: do this with the nix ioctl
     use libc::ioctl;
     use libc::TIOCGWINSZ;
@@ -54,7 +53,10 @@ pub(crate) fn get_terminal_size_using_fd(fd: RawFd) -> PositionAndSize {
     unsafe {
         ioctl(fd, TIOCGWINSZ.into(), &mut winsize)
     };
-    PositionAndSize::from(winsize)
+    Size {
+        rows: winsize.ws_row as usize,
+        cols: winsize.ws_col as usize,
+    }
 }
 
 #[derive(Clone)]
@@ -69,7 +71,7 @@ pub struct ClientOsInputOutput {
 /// Zellij client requires.
 pub trait ClientOsApi: Send + Sync {
     /// Returns the size of the terminal associated to file descriptor `fd`.
-    fn get_terminal_size_using_fd(&self, fd: RawFd) -> PositionAndSize;
+    fn get_terminal_size_using_fd(&self, fd: RawFd) -> Size;
     /// Set the terminal associated to file descriptor `fd` to
     /// [raw mode](https://en.wikipedia.org/wiki/Terminal_mode).
     fn set_raw_mode(&mut self, fd: RawFd);
@@ -94,11 +96,11 @@ pub trait ClientOsApi: Send + Sync {
     fn enable_mouse(&self);
     fn disable_mouse(&self);
     // Repeatedly send action, until stdin is readable again
-    fn start_action_repeater(&mut self, action: Action);
+    fn stdin_poller(&self) -> StdinPoller;
 }
 
 impl ClientOsApi for ClientOsInputOutput {
-    fn get_terminal_size_using_fd(&self, fd: RawFd) -> PositionAndSize {
+    fn get_terminal_size_using_fd(&self, fd: RawFd) -> Size {
         get_terminal_size_using_fd(fd)
     }
     fn set_raw_mode(&mut self, fd: RawFd) {
@@ -201,16 +203,8 @@ impl ClientOsApi for ClientOsInputOutput {
         }
     }
 
-    fn start_action_repeater(&mut self, action: Action) {
-        let mut poller = StdinPoller::default();
-
-        loop {
-            let ready = poller.ready();
-            if ready {
-                break;
-            }
-            self.send_to_server(ClientToServerMsg::Action(action.clone()));
-        }
+    fn stdin_poller(&self) -> StdinPoller {
+        StdinPoller::default()
     }
 }
 
@@ -234,7 +228,7 @@ pub fn get_client_os_input() -> Result<ClientOsInputOutput, nix::Error> {
 
 pub const DEFAULT_STDIN_POLL_TIMEOUT_MS: u64 = 10;
 
-struct StdinPoller {
+pub struct StdinPoller {
     poll: Poll,
     events: Events,
     timeout: time::Duration,
@@ -242,7 +236,7 @@ struct StdinPoller {
 
 impl StdinPoller {
     // use mio poll to check if stdin is readable without blocking
-    fn ready(&mut self) -> bool {
+    pub fn ready(&mut self) -> bool {
         self.poll
             .poll(&mut self.events, Some(self.timeout))
             .expect("could not poll stdin for readiness");
