@@ -10,39 +10,21 @@ use std::{
 use zellij_utils::{position::Position, vte, zellij_tile};
 
 const TABSTOP_WIDTH: usize = 8; // TODO: is this always right?
-const SCROLL_BACK: usize = 10_000;
+pub const SCROLL_BACK: usize = 10_000;
+pub const MAX_TITLE_STACK_SIZE: usize = 1000;
 
 use vte::{Params, Perform};
 use zellij_tile::data::{Palette, PaletteColor};
 use zellij_utils::{consts::VERSION, logging::debug_log_to_file, shared::version_number};
 
+use crate::panes::alacritty_functions::{parse_number, xparse_color};
 use crate::panes::terminal_character::{
-    CharacterStyles, CharsetIndex, Cursor, CursorShape, StandardCharset, TerminalCharacter,
-    EMPTY_TERMINAL_CHARACTER,
+    AnsiCode, CharacterStyles, CharsetIndex, Cursor, CursorShape, StandardCharset,
+    TerminalCharacter, EMPTY_TERMINAL_CHARACTER,
 };
 
 use super::selection::Selection;
 
-// this was copied verbatim from alacritty
-fn parse_number(input: &[u8]) -> Option<u8> {
-    if input.is_empty() {
-        return None;
-    }
-    let mut num: u8 = 0;
-    for c in input {
-        let c = *c as char;
-        if let Some(digit) = c.to_digit(10) {
-            num = match num.checked_mul(10).and_then(|v| v.checked_add(digit as u8)) {
-                Some(v) => v,
-                None => return None,
-            }
-        } else {
-            return None;
-        }
-    }
-    Some(num)
-}
-
 fn get_top_non_canonical_rows(rows: &mut Vec<Row>) -> Vec<Row> {
     let mut index_of_last_non_canonical_row = None;
     for (i, row) in rows.iter().enumerate() {
@@ -60,7 +42,7 @@ fn get_top_non_canonical_rows(rows: &mut Vec<Row>) -> Vec<Row> {
     }
 }
 
-fn get_bottom_canonical_row_and_wraps(rows: &mut VecDeque<Row>) -> Vec<Row> {
+fn get_lines_above_bottom_canonical_row_and_wraps(rows: &mut VecDeque<Row>) -> Vec<Row> {
     let mut index_of_last_non_canonical_row = None;
     for (i, row) in rows.iter().enumerate().rev() {
         index_of_last_non_canonical_row = Some(i);
@@ -76,100 +58,157 @@ fn get_bottom_canonical_row_and_wraps(rows: &mut VecDeque<Row>) -> Vec<Row> {
     }
 }
 
-fn transfer_rows_down(
-    source: &mut VecDeque<Row>,
-    destination: &mut Vec<Row>,
+fn get_viewport_bottom_canonical_row_and_wraps(viewport: &mut Vec<Row>) -> Vec<Row> {
+    let mut index_of_last_non_canonical_row = None;
+    for (i, row) in viewport.iter().enumerate().rev() {
+        index_of_last_non_canonical_row = Some(i);
+        if row.is_canonical {
+            break;
+        }
+    }
+    match index_of_last_non_canonical_row {
+        Some(index_of_last_non_canonical_row) => {
+            viewport.drain(index_of_last_non_canonical_row..).collect()
+        }
+        None => vec![],
+    }
+}
+
+fn get_top_canonical_row_and_wraps(rows: &mut Vec<Row>) -> Vec<Row> {
+    let mut index_of_first_non_canonical_row = None;
+    let mut end_index_of_first_canonical_line = None;
+    for (i, row) in rows.iter().enumerate() {
+        if row.is_canonical && end_index_of_first_canonical_line.is_none() {
+            index_of_first_non_canonical_row = Some(i);
+            end_index_of_first_canonical_line = Some(i);
+            continue;
+        }
+        if row.is_canonical && end_index_of_first_canonical_line.is_some() {
+            break;
+        }
+        if index_of_first_non_canonical_row.is_some() {
+            end_index_of_first_canonical_line = Some(i);
+            continue;
+        }
+    }
+    match (
+        index_of_first_non_canonical_row,
+        end_index_of_first_canonical_line,
+    ) {
+        (Some(first_index), Some(last_index)) => rows.drain(first_index..=last_index).collect(),
+        (Some(first_index), None) => rows.drain(first_index..).collect(),
+        _ => vec![],
+    }
+}
+
+fn transfer_rows_from_lines_above_to_viewport(
+    lines_above: &mut VecDeque<Row>,
+    viewport: &mut Vec<Row>,
     count: usize,
-    max_src_width: Option<usize>,
-    max_dst_width: Option<usize>,
+    max_viewport_width: usize,
 ) {
     let mut next_lines: Vec<Row> = vec![];
-    let mut lines_added_to_destination: isize = 0;
+    let mut lines_added_to_viewport: isize = 0;
     loop {
-        if lines_added_to_destination as usize == count {
+        if lines_added_to_viewport as usize == count {
             break;
         }
         if next_lines.is_empty() {
-            match source.pop_back() {
+            match lines_above.pop_back() {
                 Some(next_line) => {
-                    let mut top_non_canonical_rows_in_dst = get_top_non_canonical_rows(destination);
-                    lines_added_to_destination -= top_non_canonical_rows_in_dst.len() as isize;
+                    let mut top_non_canonical_rows_in_dst = get_top_non_canonical_rows(viewport);
+                    lines_added_to_viewport -= top_non_canonical_rows_in_dst.len() as isize;
                     next_lines.push(next_line);
                     next_lines.append(&mut top_non_canonical_rows_in_dst);
-                    next_lines = match max_dst_width {
-                        Some(max_row_width) => Row::from_rows(next_lines, max_row_width)
-                            .split_to_rows_of_length(max_row_width),
-                        None => vec![Row::from_rows(next_lines, 0)],
-                    };
+                    next_lines = Row::from_rows(next_lines, max_viewport_width)
+                        .split_to_rows_of_length(max_viewport_width);
                     if next_lines.is_empty() {
-                        // no more lines at source, the line we popped was probably empty
+                        // no more lines at lines_above, the line we popped was probably empty
                         break;
                     }
                 }
                 None => break, // no more rows
             }
         }
-        destination.insert(0, next_lines.pop().unwrap());
-        lines_added_to_destination += 1;
+        viewport.insert(0, next_lines.pop().unwrap());
+        lines_added_to_viewport += 1;
     }
     if !next_lines.is_empty() {
-        match max_src_width {
-            Some(max_row_width) => {
-                let excess_rows = Row::from_rows(next_lines, max_row_width)
-                    .split_to_rows_of_length(max_row_width);
-                source.extend(excess_rows);
-            }
-            None => {
-                let excess_row = Row::from_rows(next_lines, 0);
-                bounded_push(source, excess_row);
-            }
-        }
+        let excess_row = Row::from_rows(next_lines, 0);
+        bounded_push(lines_above, excess_row);
     }
 }
 
-fn transfer_rows_up(
-    source: &mut Vec<Row>,
-    destination: &mut VecDeque<Row>,
+fn transfer_rows_from_viewport_to_lines_above(
+    viewport: &mut Vec<Row>,
+    lines_above: &mut VecDeque<Row>,
     count: usize,
-    max_src_width: Option<usize>,
-    max_dst_width: Option<usize>,
+    max_viewport_width: usize,
 ) {
     let mut next_lines: Vec<Row> = vec![];
     for _ in 0..count {
         if next_lines.is_empty() {
-            if !source.is_empty() {
-                let next_line = source.remove(0);
+            if !viewport.is_empty() {
+                let next_line = viewport.remove(0);
                 if !next_line.is_canonical {
                     let mut bottom_canonical_row_and_wraps_in_dst =
-                        get_bottom_canonical_row_and_wraps(destination);
+                        get_lines_above_bottom_canonical_row_and_wraps(lines_above);
                     next_lines.append(&mut bottom_canonical_row_and_wraps_in_dst);
                 }
                 next_lines.push(next_line);
-                next_lines = match max_dst_width {
-                    Some(max_row_width) => Row::from_rows(next_lines, max_row_width)
-                        .split_to_rows_of_length(max_row_width),
-                    None => vec![Row::from_rows(next_lines, 0)],
-                };
+                next_lines = vec![Row::from_rows(next_lines, 0)];
             } else {
                 break; // no more rows
             }
         }
-        bounded_push(destination, next_lines.remove(0));
+        bounded_push(lines_above, next_lines.remove(0));
     }
     if !next_lines.is_empty() {
-        match max_src_width {
-            Some(max_row_width) => {
-                let excess_rows = Row::from_rows(next_lines, max_row_width)
-                    .split_to_rows_of_length(max_row_width);
-                for row in excess_rows {
-                    source.insert(0, row);
+        let excess_rows = Row::from_rows(next_lines, max_viewport_width)
+            .split_to_rows_of_length(max_viewport_width);
+        for row in excess_rows {
+            viewport.insert(0, row);
+        }
+    }
+}
+
+fn transfer_rows_from_lines_below_to_viewport(
+    lines_below: &mut Vec<Row>,
+    viewport: &mut Vec<Row>,
+    count: usize,
+    max_viewport_width: usize,
+) {
+    let mut next_lines: Vec<Row> = vec![];
+    for _ in 0..count {
+        let mut lines_pulled_from_viewport = 0;
+        if next_lines.is_empty() {
+            if !lines_below.is_empty() {
+                let mut top_non_canonical_rows_in_lines_below =
+                    get_top_non_canonical_rows(lines_below);
+                if !top_non_canonical_rows_in_lines_below.is_empty() {
+                    let mut canonical_line = get_viewport_bottom_canonical_row_and_wraps(viewport);
+                    lines_pulled_from_viewport += canonical_line.len();
+                    canonical_line.append(&mut top_non_canonical_rows_in_lines_below);
+                    next_lines = Row::from_rows(canonical_line, max_viewport_width)
+                        .split_to_rows_of_length(max_viewport_width);
+                } else {
+                    let canonical_row = get_top_canonical_row_and_wraps(lines_below);
+                    next_lines = Row::from_rows(canonical_row, max_viewport_width)
+                        .split_to_rows_of_length(max_viewport_width);
                 }
-            }
-            None => {
-                let excess_row = Row::from_rows(next_lines, 0);
-                source.insert(0, excess_row);
+            } else {
+                break; // no more rows
             }
         }
+        for _ in 0..(lines_pulled_from_viewport + 1) {
+            if !next_lines.is_empty() {
+                viewport.push(next_lines.remove(0));
+            }
+        }
+    }
+    if !next_lines.is_empty() {
+        let excess_row = Row::from_rows(next_lines, 0);
+        lines_below.insert(0, excess_row);
     }
 }
 
@@ -308,9 +347,12 @@ pub struct Grid {
     preceding_char: Option<TerminalCharacter>,
     colors: Palette,
     output_buffer: OutputBuffer,
+    title_stack: Vec<String>,
+    pub changed_colors: Option<[Option<AnsiCode>; 256]>,
     pub should_render: bool,
     pub cursor_key_mode: bool, // DECCKM - when set, cursor keys should send ANSI direction codes (eg. "OD") instead of the arrow keys (eg. "[D")
-    pub erasure_mode: bool,    // ERM
+    pub bracketed_paste_mode: bool, // when set, paste instructions to the terminal should be escaped with a special sequence
+    pub erasure_mode: bool,         // ERM
     pub insert_mode: bool,
     pub disable_linewrap: bool,
     pub clear_viewport_before_rendering: bool,
@@ -318,6 +360,7 @@ pub struct Grid {
     pub height: usize,
     pub pending_messages_to_pty: Vec<Vec<u8>>,
     pub selection: Selection,
+    pub title: Option<String>,
 }
 
 impl Debug for Grid {
@@ -348,6 +391,7 @@ impl Grid {
             height: rows,
             should_render: true,
             cursor_key_mode: false,
+            bracketed_paste_mode: false,
             erasure_mode: false,
             insert_mode: false,
             disable_linewrap: false,
@@ -358,6 +402,9 @@ impl Grid {
             colors,
             output_buffer: Default::default(),
             selection: Default::default(),
+            title_stack: vec![],
+            title: None,
+            changed_colors: None,
         }
     }
     pub fn render_full_viewport(&mut self) {
@@ -404,6 +451,23 @@ impl Grid {
     pub fn cursor_shape(&self) -> CursorShape {
         self.cursor.get_shape()
     }
+    pub fn scrollback_position_and_length(&self) -> (usize, usize) {
+        // (position, length)
+        let mut scrollback_buffer_count = 0;
+        for row in self.lines_above.iter() {
+            let row_width = row.width();
+            // rows in lines_above are unwrapped, so we need to account for that
+            if row_width > self.width {
+                scrollback_buffer_count += (row_width as f64 / self.width as f64).ceil() as usize;
+            } else {
+                scrollback_buffer_count += 1;
+            }
+        }
+        (
+            self.lines_below.len(),
+            (scrollback_buffer_count + self.lines_below.len()),
+        )
+    }
     fn set_horizontal_tabstop(&mut self) {
         self.horizontal_tabstops.insert(self.cursor.x);
     }
@@ -463,8 +527,8 @@ impl Grid {
         for (i, line) in self.viewport.iter().enumerate() {
             if line.is_canonical {
                 canonical_lines_traversed += 1;
+                y_coordinates = i;
                 if canonical_lines_traversed == canonical_line_index + 1 {
-                    y_coordinates = i;
                     break;
                 }
             }
@@ -475,8 +539,14 @@ impl Grid {
         if !self.lines_above.is_empty() && self.viewport.len() == self.height {
             let line_to_push_down = self.viewport.pop().unwrap();
             self.lines_below.insert(0, line_to_push_down);
-            let line_to_insert_at_viewport_top = self.lines_above.pop_back().unwrap();
-            self.viewport.insert(0, line_to_insert_at_viewport_top);
+
+            transfer_rows_from_lines_above_to_viewport(
+                &mut self.lines_above,
+                &mut self.viewport,
+                1,
+                self.width,
+            );
+
             self.selection.move_down(1);
         }
         self.output_buffer.update_all_lines();
@@ -491,15 +561,44 @@ impl Grid {
                 last_line_above.append(&mut line_to_push_up.columns);
                 bounded_push(&mut self.lines_above, last_line_above);
             }
-            let line_to_insert_at_viewport_bottom = self.lines_below.remove(0);
-            self.viewport.push(line_to_insert_at_viewport_bottom);
+
+            transfer_rows_from_lines_below_to_viewport(
+                &mut self.lines_below,
+                &mut self.viewport,
+                1,
+                self.width,
+            );
+
             self.selection.move_up(1);
             self.output_buffer.update_all_lines();
         }
     }
+    fn force_change_size(&mut self, new_rows: usize, new_columns: usize) {
+        // this is an ugly hack - it's here because sometimes we need to change_size to the
+        // existing size (eg. when resizing an alternative_grid to the current height/width) and
+        // the change_size method is a no-op in that case. Should be fixed by making the
+        // change_size method atomic
+        let intermediate_rows = if new_rows == self.height {
+            new_rows + 1
+        } else {
+            new_rows
+        };
+        let intermediate_columns = if new_columns == self.width {
+            new_columns + 1
+        } else {
+            new_columns
+        };
+        self.change_size(intermediate_rows, intermediate_columns);
+        self.change_size(new_rows, new_columns);
+    }
     pub fn change_size(&mut self, new_rows: usize, new_columns: usize) {
+        // Do nothing if this pane hasn't been given a proper size yet
+        if new_columns == 0 || new_rows == 0 {
+            return;
+        }
         self.selection.reset();
         if new_columns != self.width {
+            self.horizontal_tabstops = create_horizontal_tabstops(new_columns);
             let mut cursor_canonical_line_index = self.cursor_canonical_line_index();
             let cursor_index_in_canonical_line = self.cursor_index_in_canonical_line();
             let mut viewport_canonical_lines = vec![];
@@ -529,6 +628,23 @@ impl Grid {
                     }
                 }
             }
+
+            // trim lines after the last empty space that has no following character, because
+            // terminals don't trim empty lines
+            for line in viewport_canonical_lines.iter_mut() {
+                let mut trim_at = None;
+                for (index, character) in line.columns.iter().enumerate() {
+                    if character.character != EMPTY_TERMINAL_CHARACTER.character {
+                        trim_at = None;
+                    } else if trim_at.is_none() {
+                        trim_at = Some(index);
+                    }
+                }
+                if let Some(trim_at) = trim_at {
+                    line.columns.truncate(trim_at);
+                }
+            }
+
             let mut new_viewport_rows = vec![];
             for mut canonical_line in viewport_canonical_lines {
                 let mut canonical_line_parts: Vec<Row> = vec![];
@@ -541,6 +657,11 @@ impl Grid {
                     } else {
                         canonical_line.columns.drain(..).collect()
                     };
+                    // If the next character is wider than the grid (i.e. there is nothing in
+                    // `next_wrap`, then just abort the resizing
+                    if next_wrap.is_empty() {
+                        break;
+                    }
                     let row = Row::from_columns(next_wrap);
                     // if there are no more parts, this row is canonical as long as it originally
                     // was canonical (it might not have been for example if it's the first row in
@@ -554,21 +675,23 @@ impl Grid {
                 }
                 new_viewport_rows.append(&mut canonical_line_parts);
             }
+
             self.viewport = new_viewport_rows;
 
             let mut new_cursor_y = self.canonical_line_y_coordinates(cursor_canonical_line_index);
+
             let new_cursor_x = (cursor_index_in_canonical_line / new_columns)
                 + (cursor_index_in_canonical_line % new_columns);
             let current_viewport_row_count = self.viewport.len();
             match current_viewport_row_count.cmp(&self.height) {
                 Ordering::Less => {
                     let row_count_to_transfer = self.height - current_viewport_row_count;
-                    transfer_rows_down(
+
+                    transfer_rows_from_lines_above_to_viewport(
                         &mut self.lines_above,
                         &mut self.viewport,
                         row_count_to_transfer,
-                        None,
-                        Some(new_columns),
+                        new_columns,
                     );
                     let rows_pulled = self.viewport.len() - current_viewport_row_count;
                     new_cursor_y += rows_pulled;
@@ -580,12 +703,11 @@ impl Grid {
                     } else {
                         new_cursor_y -= row_count_to_transfer;
                     }
-                    transfer_rows_up(
+                    transfer_rows_from_viewport_to_lines_above(
                         &mut self.viewport,
                         &mut self.lines_above,
                         row_count_to_transfer,
-                        Some(new_columns),
-                        None,
+                        new_columns,
                     );
                 }
                 Ordering::Equal => {}
@@ -598,12 +720,11 @@ impl Grid {
             match current_viewport_row_count.cmp(&new_rows) {
                 Ordering::Less => {
                     let row_count_to_transfer = new_rows - current_viewport_row_count;
-                    transfer_rows_down(
+                    transfer_rows_from_lines_above_to_viewport(
                         &mut self.lines_above,
                         &mut self.viewport,
                         row_count_to_transfer,
-                        None,
-                        Some(new_columns),
+                        new_columns,
                     );
                     let rows_pulled = self.viewport.len() - current_viewport_row_count;
                     self.cursor.y += rows_pulled;
@@ -615,12 +736,11 @@ impl Grid {
                     } else {
                         self.cursor.y -= row_count_to_transfer;
                     }
-                    transfer_rows_up(
+                    transfer_rows_from_viewport_to_lines_above(
                         &mut self.viewport,
                         &mut self.lines_above,
                         row_count_to_transfer,
-                        Some(new_columns),
-                        None,
+                        new_columns,
                     );
                 }
                 Ordering::Equal => {}
@@ -723,7 +843,13 @@ impl Grid {
         }
     }
     pub fn fill_viewport(&mut self, character: TerminalCharacter) {
-        self.viewport.clear();
+        let row_count_to_transfer = self.viewport.len();
+        transfer_rows_from_viewport_to_lines_above(
+            &mut self.viewport,
+            &mut self.lines_above,
+            row_count_to_transfer,
+            self.width,
+        );
         for _ in 0..self.height {
             let columns = vec![character; self.width];
             self.viewport.push(Row::from_columns(columns).canonical());
@@ -763,15 +889,16 @@ impl Grid {
             self.viewport.push(new_row);
         }
         if self.cursor.y == self.height - 1 {
-            let row_count_to_transfer = 1;
-            transfer_rows_up(
-                &mut self.viewport,
-                &mut self.lines_above,
-                row_count_to_transfer,
-                Some(self.width),
-                None,
-            );
-            self.selection.move_up(1);
+            if self.scroll_region.is_none() {
+                let row_count_to_transfer = 1;
+                transfer_rows_from_viewport_to_lines_above(
+                    &mut self.viewport,
+                    &mut self.lines_above,
+                    row_count_to_transfer,
+                    self.width,
+                );
+                self.selection.move_up(1);
+            }
             self.output_buffer.update_all_lines();
         } else {
             self.cursor.y += 1;
@@ -839,12 +966,11 @@ impl Grid {
             self.cursor.x = 0;
             if self.cursor.y == self.height - 1 {
                 let row_count_to_transfer = 1;
-                transfer_rows_up(
+                transfer_rows_from_viewport_to_lines_above(
                     &mut self.viewport,
                     &mut self.lines_above,
                     row_count_to_transfer,
-                    Some(self.width),
-                    None,
+                    self.width,
                 );
                 let wrapped_row = Row::new(self.width);
                 self.viewport.push(wrapped_row);
@@ -867,10 +993,9 @@ impl Grid {
         self.cursor.x += count_to_move;
     }
     pub fn replace_characters_in_line_after_cursor(&mut self, replace_with: TerminalCharacter) {
-        self.viewport
-            .get_mut(self.cursor.y)
-            .unwrap()
-            .replace_and_pad_end(self.cursor.x, self.width, replace_with);
+        if let Some(row) = self.viewport.get_mut(self.cursor.y) {
+            row.replace_and_pad_end(self.cursor.x, self.width, replace_with);
+        }
         self.output_buffer.update_line(self.cursor.y);
     }
     pub fn replace_characters_in_line_before_cursor(&mut self, replace_with: TerminalCharacter) {
@@ -934,7 +1059,11 @@ impl Grid {
                 } else {
                     0
                 };
-                self.cursor.y = std::cmp::min(scroll_region_bottom, y + y_offset);
+                if y >= scroll_region_top && y <= scroll_region_bottom {
+                    self.cursor.y = std::cmp::min(scroll_region_bottom, y + y_offset);
+                } else {
+                    self.cursor.y = std::cmp::min(self.height - 1, y + y_offset);
+                }
                 self.pad_lines_until(self.cursor.y, pad_character);
                 self.pad_current_line_until(self.cursor.x);
             }
@@ -981,26 +1110,18 @@ impl Grid {
         }
         self.output_buffer.update_all_lines();
     }
-    pub fn move_cursor_down(&mut self, count: usize, pad_character: TerminalCharacter) {
+    pub fn move_cursor_down_until_edge_of_screen(
+        &mut self,
+        count: usize,
+        pad_character: TerminalCharacter,
+    ) {
         if let Some((scroll_region_top, scroll_region_bottom)) = self.scroll_region {
             if self.cursor.y >= scroll_region_top && self.cursor.y <= scroll_region_bottom {
                 self.cursor.y = std::cmp::min(self.cursor.y + count, scroll_region_bottom);
                 return;
             }
         }
-        let lines_to_add = if self.cursor.y + count > self.height - 1 {
-            (self.cursor.y + count) - (self.height - 1)
-        } else {
-            0
-        };
-        self.cursor.y = if self.cursor.y + count > self.height - 1 {
-            self.height - 1
-        } else {
-            self.cursor.y + count
-        };
-        for _ in 0..lines_to_add {
-            self.add_canonical_line();
-        }
+        self.cursor.y = std::cmp::min(self.cursor.y + count, self.height - 1);
         self.pad_lines_until(self.cursor.y, pad_character);
     }
     pub fn move_cursor_back(&mut self, count: usize) {
@@ -1028,7 +1149,7 @@ impl Grid {
         self.scroll_region = None;
     }
     pub fn set_scroll_region_to_viewport_size(&mut self) {
-        self.scroll_region = Some((0, self.height - 1));
+        self.scroll_region = Some((0, self.height.saturating_sub(1)));
     }
     pub fn delete_lines_in_scroll_region(
         &mut self,
@@ -1138,8 +1259,9 @@ impl Grid {
         self.active_charset = Default::default();
         self.erasure_mode = false;
         self.disable_linewrap = false;
-        self.cursor.change_shape(CursorShape::Block);
+        self.cursor.change_shape(CursorShape::Initial);
         self.output_buffer.update_all_lines();
+        self.changed_colors = None;
     }
     fn set_preceding_character(&mut self, terminal_character: TerminalCharacter) {
         self.preceding_char = Some(terminal_character);
@@ -1200,7 +1322,7 @@ impl Grid {
             let empty_row = Row::from_columns(vec![EMPTY_TERMINAL_CHARACTER; self.width]);
 
             // get the row from lines_above, viewport, or lines below depending on index
-            let row = if l < 0 {
+            let row = if l < 0 && self.lines_above.len() > l.abs() as usize {
                 let offset_from_end = l.abs();
                 &self.lines_above[self
                     .lines_above
@@ -1211,8 +1333,12 @@ impl Grid {
             } else if (l as usize) < self.height {
                 // index is in viewport but there is no line
                 &empty_row
-            } else {
+            } else if self.lines_below.len() > (l as usize).saturating_sub(self.viewport.len()) {
                 &self.lines_below[(l as usize) - self.viewport.len()]
+            } else {
+                // can't find the line, this probably it's on the pane border
+                // is on the pane border
+                continue;
             };
 
             let excess_width = row.excess_width();
@@ -1242,11 +1368,28 @@ impl Grid {
             self.output_buffer.update_line(l as usize);
         }
     }
+    fn set_title(&mut self, title: String) {
+        self.title = Some(title);
+    }
+    fn push_current_title_to_stack(&mut self) {
+        if self.title_stack.len() > MAX_TITLE_STACK_SIZE {
+            self.title_stack.remove(0);
+        }
+        if let Some(title) = self.title.as_ref() {
+            self.title_stack.push(title.clone());
+        }
+    }
+    fn pop_title_from_stack(&mut self) {
+        if let Some(popped_title) = self.title_stack.pop() {
+            self.title = Some(popped_title);
+        }
+    }
 }
 
 impl Perform for Grid {
     fn print(&mut self, c: char) {
         let c = self.cursor.charsets[self.active_charset].map(c);
+
         // apparently, building TerminalCharacter like this without a "new" method
         // is a little faster
         let terminal_character = TerminalCharacter {
@@ -1311,24 +1454,30 @@ impl Perform for Grid {
             // Set window title.
             b"0" | b"2" => {
                 if params.len() >= 2 {
-                    let _title = params[1..]
+                    let title = params[1..]
                         .iter()
                         .flat_map(|x| str::from_utf8(x))
                         .collect::<Vec<&str>>()
                         .join(";")
                         .trim()
                         .to_owned();
-                    // TBD: do something with title?
+                    self.set_title(title);
                 }
             }
 
             // Set color index.
             b"4" => {
-                // TBD: set color index - currently unsupported
-                //
-                // this changes a terminal color index to something else
-                // meaning anything set to that index will be changed
-                // during rendering
+                for chunk in params[1..].chunks(2) {
+                    let index = parse_number(chunk[0]);
+                    let color = xparse_color(chunk[1]);
+                    if let (Some(i), Some(c)) = (index, color) {
+                        if self.changed_colors.is_none() {
+                            self.changed_colors = Some([None; 256]);
+                        }
+                        self.changed_colors.as_mut().unwrap()[i as usize] = Some(c);
+                        return;
+                    }
+                }
             }
 
             // Get/set Foreground, Background, Cursor colors.
@@ -1402,6 +1551,21 @@ impl Perform for Grid {
 
             // Reset color index.
             b"104" => {
+                // Reset all color indexes when no parameters are given.
+                if params.len() == 1 {
+                    self.changed_colors = None;
+                    return;
+                }
+
+                // Reset color indexes given as parameters.
+                for param in &params[1..] {
+                    if let Some(index) = parse_number(param) {
+                        if self.changed_colors.is_some() {
+                            self.changed_colors.as_mut().unwrap()[index as usize] = None
+                        }
+                    }
+                }
+
                 // Reset all color indexes when no parameters are given.
                 if params.len() == 1 {
                     // TBD - reset all color changes - currently unsupported
@@ -1496,7 +1660,7 @@ impl Perform for Grid {
             // move cursor down until edge of screen
             let move_down_count = next_param_or(1);
             let pad_character = EMPTY_TERMINAL_CHARACTER;
-            self.move_cursor_down(move_down_count as usize, pad_character);
+            self.move_cursor_down_until_edge_of_screen(move_down_count as usize, pad_character);
         } else if c == 'D' {
             let move_back_count = next_param_or(1);
             self.move_cursor_back(move_back_count);
@@ -1508,6 +1672,9 @@ impl Perform for Grid {
             };
             if first_intermediate_is_questionmark {
                 match params_iter.next().map(|param| param[0]) {
+                    Some(2004) => {
+                        self.bracketed_paste_mode = false;
+                    }
                     Some(1049) => {
                         if let Some((
                             alternative_lines_above,
@@ -1521,7 +1688,7 @@ impl Perform for Grid {
                         }
                         self.alternative_lines_above_viewport_and_cursor = None;
                         self.clear_viewport_before_rendering = true;
-                        self.change_size(self.height, self.width); // the alternative_viewport might have been of a different size...
+                        self.force_change_size(self.height, self.width); // the alternative_viewport might have been of a different size...
                         self.mark_for_rerender();
                     }
                     Some(25) => {
@@ -1561,6 +1728,9 @@ impl Perform for Grid {
                         self.show_cursor();
                         self.mark_for_rerender();
                     }
+                    Some(2004) => {
+                        self.bracketed_paste_mode = true;
+                    }
                     Some(1049) => {
                         let current_lines_above = std::mem::replace(
                             &mut self.lines_above,
@@ -1676,7 +1846,7 @@ impl Perform for Grid {
         } else if c == 'E' {
             let count = next_param_or(1);
             let pad_character = EMPTY_TERMINAL_CHARACTER;
-            self.move_cursor_down(count, pad_character);
+            self.move_cursor_down_until_edge_of_screen(count, pad_character);
         } else if c == 'F' {
             let count = next_param_or(1);
             self.move_cursor_up(count);
@@ -1691,7 +1861,8 @@ impl Perform for Grid {
                 // DECSCUSR (CSI Ps SP q) -- Set Cursor Style.
                 let cursor_style_id = next_param_or(0);
                 let shape = match cursor_style_id {
-                    0 | 2 => Some(CursorShape::Block),
+                    0 => Some(CursorShape::Initial),
+                    2 => Some(CursorShape::Block),
                     1 => Some(CursorShape::BlinkingBlock),
                     3 => Some(CursorShape::BlinkingUnderline),
                     4 => Some(CursorShape::Underline),
@@ -1758,10 +1929,10 @@ impl Perform for Grid {
                         .push(text_area_report.as_bytes().to_vec());
                 }
                 22 => {
-                    // TODO: push title
+                    self.push_current_title_to_stack();
                 }
                 23 => {
-                    // TODO: pop title
+                    self.pop_title_from_stack();
                 }
                 _ => {}
             }
@@ -2085,6 +2256,9 @@ impl Row {
         if !parts.is_empty() && self.is_canonical {
             parts.get_mut(0).unwrap().is_canonical = true;
         }
+        if parts.is_empty() {
+            parts.push(self.clone());
+        }
         parts
     }
 }
