@@ -1,9 +1,11 @@
+mod alacritty_functions;
 mod grid;
 mod plugin_pane;
 mod selection;
 mod terminal_character;
 mod terminal_pane;
 
+pub use alacritty_functions::*;
 pub use grid::*;
 pub(crate) use plugin_pane::*;
 pub use terminal_character::*;
