@@ -13,14 +13,17 @@ use zellij_utils::{
         command::TerminalAction,
         get_mode_info,
     },
-    ipc::{ClientToServerMsg, ExitReason, ServerToClientMsg},
+    ipc::{ClientToServerMsg, IpcReceiverWithContext, ServerToClientMsg},
 };
 
+use crate::ClientId;
+
 fn route_action(
     action: Action,
     session: &SessionMetaData,
     _os_input: &dyn ServerOsApi,
     to_server: &SenderWithContext<ServerInstruction>,
+    client_id: ClientId,
 ) -> bool {
     let mut should_break = false;
     session
@@ -241,11 +244,15 @@ fn route_action(
                 .unwrap();
         }
         Action::Quit => {
-            to_server.send(ServerInstruction::ClientExit).unwrap();
+            to_server
+                .send(ServerInstruction::ClientExit(client_id))
+                .unwrap();
             should_break = true;
         }
         Action::Detach => {
-            to_server.send(ServerInstruction::DetachSession).unwrap();
+            to_server
+                .send(ServerInstruction::DetachSession(client_id))
+                .unwrap();
             should_break = true;
         }
         Action::LeftClick(point) => {
@@ -282,47 +289,75 @@ pub(crate) fn route_thread_main(
     session_state: Arc<RwLock<SessionState>>,
     os_input: Box<dyn ServerOsApi>,
     to_server: SenderWithContext<ServerInstruction>,
+    mut receiver: IpcReceiverWithContext<ClientToServerMsg>,
+    client_id: ClientId,
 ) {
     loop {
-        let (instruction, err_ctx) = os_input.recv_from_client();
+        let (instruction, err_ctx) = receiver.recv();
         err_ctx.update_thread_ctx();
         let rlocked_sessions = session_data.read().unwrap();
 
         match instruction {
             ClientToServerMsg::Action(action) => {
                 if let Some(rlocked_sessions) = rlocked_sessions.as_ref() {
-                    if route_action(action, rlocked_sessions, &*os_input, &to_server) {
+                    if let Action::SwitchToMode(input_mode) = action {
+                        for client_id in session_state.read().unwrap().clients.keys() {
+                            os_input.send_to_client(
+                                *client_id,
+                                ServerToClientMsg::SwitchToMode(input_mode),
+                            );
+                        }
+                    }
+                    if route_action(action, rlocked_sessions, &*os_input, &to_server, client_id) {
                         break;
                     }
                 }
             }
             ClientToServerMsg::TerminalResize(new_size) => {
+                session_state
+                    .write()
+                    .unwrap()
+                    .set_client_size(client_id, new_size);
+                let min_size = session_state
+                    .read()
+                    .unwrap()
+                    .min_client_terminal_size()
+                    .unwrap();
                 rlocked_sessions
                     .as_ref()
                     .unwrap()
                     .senders
-                    .send_to_screen(ScreenInstruction::TerminalResize(new_size))
+                    .send_to_screen(ScreenInstruction::TerminalResize(min_size))
                     .unwrap();
             }
-            ClientToServerMsg::NewClient(..) => {
-                if *session_state.read().unwrap() != SessionState::Uninitialized {
-                    os_input.send_to_temp_client(ServerToClientMsg::Exit(ExitReason::Error(
-                        "Cannot add new client".into(),
-                    )));
-                } else {
-                    os_input.add_client_sender();
-                    to_server.send(instruction.into()).unwrap();
-                }
+            ClientToServerMsg::NewClient(
+                client_attributes,
+                cli_args,
+                opts,
+                layout,
+                plugin_config,
+            ) => {
+                let new_client_instruction = ServerInstruction::NewClient(
+                    client_attributes,
+                    cli_args,
+                    opts,
+                    layout,
+                    client_id,
+                    plugin_config,
+                );
+                to_server.send(new_client_instruction).unwrap();
             }
-            ClientToServerMsg::AttachClient(_, force, _) => {
-                if *session_state.read().unwrap() == SessionState::Attached && !force {
-                    os_input.send_to_temp_client(ServerToClientMsg::Exit(ExitReason::CannotAttach));
-                } else {
-                    os_input.add_client_sender();
-                    to_server.send(instruction.into()).unwrap();
-                }
+            ClientToServerMsg::AttachClient(client_attributes, opts) => {
+                let attach_client_instruction =
+                    ServerInstruction::AttachClient(client_attributes, opts, client_id);
+                to_server.send(attach_client_instruction).unwrap();
+            }
+            ClientToServerMsg::ClientExited => {
+                // we don't unwrap this because we don't really care if there's an error here (eg.
+                // if the main server thread exited before this router thread did)
+                let _ = to_server.send(ServerInstruction::RemoveClient(client_id));
+                break;
             }
-            ClientToServerMsg::ClientExited => break,
         }
     }
 }
