@@ -1,3 +1,3 @@
 pub mod boundaries;
+pub mod pane_boundaries_frame;
 pub mod pane_resizer;
-pub mod pane_resizer_beta;
