@@ -3,7 +3,7 @@
 use crate::{
     cli::CliArgs,
     errors::{get_current_ctx, ErrorContext},
-    input::{actions::Action, layout::LayoutFromYaml, options::Options},
+    input::{actions::Action, layout::LayoutFromYaml, options::Options, plugins::PluginsConfig},
     pane_size::Size,
 };
 use interprocess::local_socket::LocalSocketStream;
@@ -16,7 +16,7 @@ use std::{
     os::unix::io::{AsRawFd, FromRawFd},
 };
 
-use zellij_tile::data::Palette;
+use zellij_tile::data::{InputMode, Palette};
 
 type SessionId = u64;
 
@@ -58,8 +58,14 @@ pub enum ClientToServerMsg {
     // Disconnect from the session we're connected to
     DisconnectFromSession,*/
     TerminalResize(Size),
-    NewClient(ClientAttributes, Box<CliArgs>, Box<Options>, LayoutFromYaml),
-    AttachClient(ClientAttributes, bool, Options),
+    NewClient(
+        ClientAttributes,
+        Box<CliArgs>,
+        Box<Options>,
+        LayoutFromYaml,
+        Option<PluginsConfig>,
+    ),
+    AttachClient(ClientAttributes, Options),
     Action(Action),
     ClientExited,
 }
@@ -74,6 +80,7 @@ pub enum ServerToClientMsg {
     Render(String),
     UnblockInputThread,
     Exit(ExitReason),
+    SwitchToMode(InputMode),
 }
 
 #[derive(Serialize, Deserialize, Debug, Clone)]
@@ -120,7 +127,9 @@ impl<T: Serialize> IpcSenderWithContext<T> {
     pub fn send(&mut self, msg: T) {
         let err_ctx = get_current_ctx();
         bincode::serialize_into(&mut self.sender, &(msg, err_ctx)).unwrap();
-        self.sender.flush().unwrap();
+        // TODO: unwrapping here can cause issues when the server disconnects which we don't mind
+        // do we need to handle errors here in other cases?
+        let _ = self.sender.flush();
     }
 
     /// Returns an [`IpcReceiverWithContext`] with the same socket as this sender.
